/***
 * Copyright (C) 1996-98. by Miroslav Zagorac <zaga@tirnanog.cc.fer.hr>
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>

#include "../lib/lib_debug.h"

#ifdef USE_SVGALIB
#  include "../lib/lib_svga.h"

#  define _svga(a,b)   a,b
#endif

#ifdef USE_X11
#  include "../lib/lib_X11.h"

#  define _x11(a,b,c)   a,b,c
#endif

#include "keyboard.h"
#include "main.h"


#ifndef _svga
#  define _svga(a,b)
#endif

#ifndef _x11
#  define _x11(a,b,c)
#endif

static struct
{
#ifdef USE_SVGALIB
   short  scancode;
#endif
   char   status;
   char   ascii;
   char   ascii_shift;
#ifdef USE_X11
   KeySym keysym;
   KeySym keysym_shift;
#endif
} keys_table[128] =
   {
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(SCANCODE_ESCAPE           ,) 0, '\033', '\033' _x11(, XK_Escape      , XK_Escape      ) },
      { _svga(SCANCODE_1                ,) 0,    '1',    '!' _x11(, XK_1           , XK_1           ) },
      { _svga(SCANCODE_2                ,) 0,    '2',    '@' _x11(, XK_2           , XK_2           ) },
      { _svga(SCANCODE_3                ,) 0,    '3',    '#' _x11(, XK_3           , XK_3           ) },
      { _svga(SCANCODE_4                ,) 0,    '4',    '$' _x11(, XK_4           , XK_4           ) },
      { _svga(SCANCODE_5                ,) 0,    '5',    '%' _x11(, XK_5           , XK_5           ) },
      { _svga(SCANCODE_6                ,) 0,    '6',    '^' _x11(, XK_6           , XK_6           ) },
      { _svga(SCANCODE_7                ,) 0,    '7',    '&' _x11(, XK_7           , XK_7           ) },
      { _svga(SCANCODE_8                ,) 0,    '8',    '*' _x11(, XK_8           , XK_8           ) },
      { _svga(SCANCODE_9                ,) 0,    '9',    '(' _x11(, XK_9           , XK_9           ) },
      { _svga(SCANCODE_0                ,) 0,    '0',    ')' _x11(, XK_0           , XK_0           ) },
      { _svga(SCANCODE_MINUS            ,) 0,    '-',    '_' _x11(, XK_minus       , XK_minus       ) },
      { _svga(SCANCODE_EQUAL            ,) 0,    '=',    '+' _x11(, XK_equal       , XK_plus        ) },
      { _svga(SCANCODE_BACKSPACE        ,) 0, '\010', '\010' _x11(, XK_BackSpace   , XK_BackSpace   ) },
      { _svga(SCANCODE_TAB              ,) 0, '\011', '\011' _x11(, XK_Tab         , XK_Tab         ) },
      { _svga(SCANCODE_Q                ,) 0,    'q',    'Q' _x11(, XK_q           , XK_Q           ) },
      { _svga(SCANCODE_W                ,) 0,    'w',    'W' _x11(, XK_w           , XK_W           ) },
      { _svga(SCANCODE_E                ,) 0,    'e',    'E' _x11(, XK_e           , XK_E           ) },
      { _svga(SCANCODE_R                ,) 0,    'r',    'R' _x11(, XK_r           , XK_R           ) },
      { _svga(SCANCODE_T                ,) 0,    't',    'T' _x11(, XK_t           , XK_T           ) },
      { _svga(SCANCODE_Y                ,) 0,    'y',    'Y' _x11(, XK_y           , XK_Y           ) },
      { _svga(SCANCODE_U                ,) 0,    'u',    'U' _x11(, XK_u           , XK_U           ) },
      { _svga(SCANCODE_I                ,) 0,    'i',    'I' _x11(, XK_i           , XK_I           ) },
      { _svga(SCANCODE_O                ,) 0,    'o',    'O' _x11(, XK_o           , XK_O           ) },
      { _svga(SCANCODE_P                ,) 0,    'p',    'P' _x11(, XK_p           , XK_P           ) },
      { _svga(SCANCODE_BRACKET_LEFT     ,) 0,    '[',    '{' _x11(, XK_bracketleft , XK_braceleft   ) },
      { _svga(SCANCODE_BRACKET_RIGHT    ,) 0,    ']',    '}' _x11(, XK_bracketright, XK_braceright  ) },
      { _svga(SCANCODE_ENTER            ,) 0, '\012', '\012' _x11(, XK_Return      , XK_Return      ) },
      { _svga(SCANCODE_LEFTCONTROL      ,) 0, '\000', '\000' _x11(, XK_Control_L   , XK_Control_L   ) },
      { _svga(SCANCODE_A                ,) 0,    'a',    'A' _x11(, XK_a           , XK_A           ) },
      { _svga(SCANCODE_S                ,) 0,    's',    'S' _x11(, XK_s           , XK_S           ) },
      { _svga(SCANCODE_D                ,) 0,    'd',    'D' _x11(, XK_d           , XK_D           ) },
      { _svga(SCANCODE_F                ,) 0,    'f',    'F' _x11(, XK_f           , XK_F           ) },
      { _svga(SCANCODE_G                ,) 0,    'g',    'G' _x11(, XK_g           , XK_G           ) },
      { _svga(SCANCODE_H                ,) 0,    'h',    'H' _x11(, XK_h           , XK_H           ) },
      { _svga(SCANCODE_J                ,) 0,    'j',    'J' _x11(, XK_j           , XK_J           ) },
      { _svga(SCANCODE_K                ,) 0,    'k',    'K' _x11(, XK_k           , XK_K           ) },
      { _svga(SCANCODE_L                ,) 0,    'l',    'L' _x11(, XK_l           , XK_L           ) },
      { _svga(SCANCODE_SEMICOLON        ,) 0,    ';',    ':' _x11(, XK_semicolon   , XK_colon       ) },
      { _svga(SCANCODE_APOSTROPHE       ,) 0,   '\'',   '\"' _x11(, XK_apostrophe  , XK_quotedbl    ) },
      { _svga(SCANCODE_GRAVE            ,) 0,    '`',    '~' _x11(, XK_grave       , XK_asciitilde  ) },
      { _svga(SCANCODE_LEFTSHIFT        ,) 0, '\000', '\000' _x11(, XK_Shift_L     , XK_Shift_L     ) },
      { _svga(SCANCODE_BACKSLASH        ,) 0,   '\\',    '|' _x11(, XK_backslash   , XK_bar         ) },
      { _svga(SCANCODE_Z                ,) 0,    'z',    'Z' _x11(, XK_z           , XK_Z           ) },
      { _svga(SCANCODE_X                ,) 0,    'x',    'X' _x11(, XK_x           , XK_X           ) },
      { _svga(SCANCODE_C                ,) 0,    'c',    'C' _x11(, XK_c           , XK_C           ) },
      { _svga(SCANCODE_V                ,) 0,    'v',    'V' _x11(, XK_v           , XK_V           ) },
      { _svga(SCANCODE_B                ,) 0,    'b',    'B' _x11(, XK_b           , XK_B           ) },
      { _svga(SCANCODE_N                ,) 0,    'n',    'N' _x11(, XK_n           , XK_N           ) },
      { _svga(SCANCODE_M                ,) 0,    'm',    'M' _x11(, XK_m           , XK_M           ) },
      { _svga(SCANCODE_COMMA            ,) 0,    ',',    '<' _x11(, XK_comma       , XK_less        ) },
      { _svga(SCANCODE_PERIOD           ,) 0,    '.',    '>' _x11(, XK_period      , XK_greater     ) },
      { _svga(SCANCODE_SLASH            ,) 0,    '/',    '?' _x11(, XK_slash       , XK_question    ) },
      { _svga(SCANCODE_RIGHTSHIFT       ,) 0, '\000', '\000' _x11(, XK_Shift_R     , XK_Shift_R     ) },
      { _svga(SCANCODE_KEYPADMULTIPLY   ,) 0,    '*',    '*' _x11(, XK_KP_Multiply , XK_KP_Multiply ) },
      { _svga(SCANCODE_LEFTALT          ,) 0, '\000', '\000' _x11(, XK_Alt_L       , XK_Alt_L       ) },
      { _svga(SCANCODE_SPACE            ,) 0,    ' ',    ' ' _x11(, XK_space       , XK_space       ) },
      { _svga(SCANCODE_CAPSLOCK         ,) 0, '\000', '\000' _x11(, XK_Caps_Lock   , XK_Caps_Lock   ) },
      { _svga(SCANCODE_F1               ,) 0, '\000', '\000' _x11(, XK_F1          , XK_F1          ) },
      { _svga(SCANCODE_F2               ,) 0, '\000', '\000' _x11(, XK_F2          , XK_F2          ) },
      { _svga(SCANCODE_F3               ,) 0, '\000', '\000' _x11(, XK_F3          , XK_F3          ) },
      { _svga(SCANCODE_F4               ,) 0, '\000', '\000' _x11(, XK_F4          , XK_F4          ) },
      { _svga(SCANCODE_F5               ,) 0, '\000', '\000' _x11(, XK_F5          , XK_F5          ) },
      { _svga(SCANCODE_F6               ,) 0, '\000', '\000' _x11(, XK_F6          , XK_F6          ) },
      { _svga(SCANCODE_F7               ,) 0, '\000', '\000' _x11(, XK_F7          , XK_F7          ) },
      { _svga(SCANCODE_F8               ,) 0, '\000', '\000' _x11(, XK_F8          , XK_F8          ) },
      { _svga(SCANCODE_F9               ,) 0, '\000', '\000' _x11(, XK_F9          , XK_F8          ) },
      { _svga(SCANCODE_F10              ,) 0, '\000', '\000' _x11(, XK_F10         , XK_F10         ) },
      { _svga(SCANCODE_NUMLOCK          ,) 0, '\000', '\000' _x11(, XK_Num_Lock    , XK_Num_Lock    ) },
      { _svga(SCANCODE_SCROLLLOCK       ,) 0, '\000', '\000' _x11(, XK_Scroll_Lock , XK_Scroll_Lock ) },
      { _svga(SCANCODE_KEYPAD7          ,) 0,    '7',    '7' _x11(, XK_KP_Home     , XK_KP_7        ) },
      { _svga(SCANCODE_KEYPAD8          ,) 0,    '8',    '8' _x11(, XK_KP_Up       , XK_KP_8        ) },
      { _svga(SCANCODE_KEYPAD9          ,) 0,    '9',    '9' _x11(, XK_KP_Prior    , XK_KP_9        ) },
      { _svga(SCANCODE_KEYPADMINUS      ,) 0,    '-',    '-' _x11(, XK_KP_Subtract , XK_KP_Subtract ) },
      { _svga(SCANCODE_KEYPAD4          ,) 0,    '4',    '4' _x11(, XK_KP_Left     , XK_KP_4        ) },
      { _svga(SCANCODE_KEYPAD5          ,) 0,    '5',    '5' _x11(, XK_KP_Begin    , XK_KP_5        ) },
      { _svga(SCANCODE_KEYPAD6          ,) 0,    '6',    '6' _x11(, XK_KP_Right    , XK_KP_6        ) },
      { _svga(SCANCODE_KEYPADPLUS       ,) 0,    '-',    '-' _x11(, XK_KP_Add      , XK_KP_Add      ) },
      { _svga(SCANCODE_KEYPAD1          ,) 0,    '1',    '1' _x11(, XK_KP_End      , XK_KP_1        ) },
      { _svga(SCANCODE_KEYPAD2          ,) 0,    '2',    '2' _x11(, XK_KP_Down     , XK_KP_2        ) },
      { _svga(SCANCODE_KEYPAD3          ,) 0,    '3',    '3' _x11(, XK_KP_Next     , XK_KP_3        ) },
      { _svga(SCANCODE_KEYPAD0          ,) 0,    '0',    '0' _x11(, XK_KP_Insert   , XK_KP_0        ) },
      { _svga(SCANCODE_KEYPADPERIOD     ,) 0,    '.',    '.' _x11(, XK_KP_Decimal  , XK_KP_Decimal  ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(SCANCODE_LESS             ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(SCANCODE_F11              ,) 0, '\000', '\000' _x11(, XK_F11         , XK_F11         ) },
      { _svga(SCANCODE_F12              ,) 0, '\000', '\000' _x11(, XK_F12         , XK_F12         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(SCANCODE_KEYPADENTER      ,) 0, '\000', '\000' _x11(, XK_KP_Enter    , XK_KP_Enter    ) },
      { _svga(SCANCODE_RIGHTCONTROL     ,) 0, '\000', '\000' _x11(, XK_Control_R   , XK_Control_R   ) },
      { _svga(SCANCODE_KEYPADDIVIDE     ,) 0, '\000', '\000' _x11(, XK_KP_Divide   , XK_KP_Divide   ) },
      { _svga(SCANCODE_PRINTSCREEN      ,) 0, '\000', '\000' _x11(, XK_Print       , XK_Print       ) },
      { _svga(SCANCODE_RIGHTALT         ,) 0, '\000', '\000' _x11(, XK_Alt_R       , XK_Alt_R       ) },
      { _svga(SCANCODE_BREAK            ,) 0, '\000', '\000' _x11(, XK_Break       , XK_Break       ) },
      { _svga(SCANCODE_HOME             ,) 0, '\000', '\000' _x11(, XK_Home        , XK_Home        ) },
      { _svga(SCANCODE_CURSORBLOCKUP    ,) 0, '\000', '\000' _x11(, XK_Up          , XK_Up          ) },
      { _svga(SCANCODE_PAGEUP           ,) 0, '\000', '\000' _x11(, XK_Page_Up     , XK_Page_Up     ) },
      { _svga(SCANCODE_CURSORBLOCKLEFT  ,) 0, '\000', '\000' _x11(, XK_Left        , XK_Left        ) },
      { _svga(SCANCODE_CURSORBLOCKRIGHT ,) 0, '\000', '\000' _x11(, XK_Right       , XK_Right       ) },
      { _svga(SCANCODE_END              ,) 0, '\000', '\000' _x11(, XK_End         , XK_End         ) },
      { _svga(SCANCODE_CURSORBLOCKDOWN  ,) 0, '\000', '\000' _x11(, XK_Down        , XK_Down        ) },
      { _svga(SCANCODE_PAGEDOWN         ,) 0, '\000', '\000' _x11(, XK_Page_Down   , XK_Page_Down   ) },
      { _svga(SCANCODE_INSERT           ,) 0, '\000', '\000' _x11(, XK_Insert      , XK_Insert      ) },
      { _svga(SCANCODE_REMOVE           ,) 0, '\000', '\000' _x11(, XK_Delete      , XK_Delete      ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(SCANCODE_BREAK_ALTERNATIVE,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
      { _svga(0                         ,) 0, '\000', '\000' _x11(, 0x0000         , 0x0000         ) },
   };

#undef _svga
#undef _x11


#ifdef USE_SVGALIB

short svga_key_define[BD_MAXKEYS][2] =
{
   /***
    * game keys
    */
   {                  0, SCANCODE_O                },	/* KEY_GAME_UP      */
   {                  0, SCANCODE_K                },	/* KEY_GAME_DOWN    */
   {                  0, SCANCODE_X                },	/* KEY_GAME_LEFT    */
   {                  0, SCANCODE_C                },	/* KEY_GAME_RIGHT   */
   {                  0, SCANCODE_0                },	/* KEY_GAME_FIRE    */
   {                  0, SCANCODE_Q                },	/* KEY_GAME_QUIT    */
   {                  0, SCANCODE_S                },	/* KEY_GAME_RESET   */
   {                  0, SCANCODE_P                },	/* KEY_GAME_PLAYERS */
   {                  0, SCANCODE_T                },	/* KEY_GAME_SCORE   */
   {                  0, SCANCODE_E                },	/* KEY_GAME_EDITOR  */

   /***
    * common keys
    */
   {                  0, SCANCODE_H                },	/* KEY_HELP         */
   {                  0, SCANCODE_SPACE            },	/* KEY_SPACE        */
   {                  0, SCANCODE_ENTER            },	/* KEY_ENTER        */
   {                  0, SCANCODE_F1               },	/* KEY_SNAPSHOT     */

   /***
    * editor keys
    */
   {                  0, SCANCODE_CURSORBLOCKUP    },	/* KEY_EDITOR_UP     */
   {                  0, SCANCODE_CURSORBLOCKDOWN  },	/* KEY_EDITOR_DOWN   */
   {                  0, SCANCODE_CURSORBLOCKLEFT  },	/* KEY_EDITOR_LEFT   */
   {                  0, SCANCODE_CURSORBLOCKRIGHT },	/* KEY_EDITOR_RIGHT  */
   { SCANCODE_LEFTSHIFT, SCANCODE_C                },	/* KEY_EDITOR_CLEAR  */
   { SCANCODE_LEFTSHIFT, SCANCODE_U                },	/* KEY_EDITOR_UNDO   */
   { SCANCODE_LEFTSHIFT, SCANCODE_L                },	/* KEY_EDITOR_LOAD   */
   { SCANCODE_LEFTSHIFT, SCANCODE_S                },	/* KEY_EDITOR_SAVE   */
   { SCANCODE_LEFTSHIFT, SCANCODE_Q                },	/* KEY_EDITOR_QUIT   */
   {                  0, SCANCODE_INSERT           },	/* KEY_EDITOR_INSERT */
   {                  0, SCANCODE_HOME             },	/* KEY_EDITOR_HOME   */
   {                  0, SCANCODE_END              },	/* KEY_EDITOR_END    */
   {                  0, SCANCODE_PAGEUP           },	/* KEY_EDITOR_TOP    */
   {                  0, SCANCODE_PAGEDOWN         },	/* KEY_EDITOR_BOTTOM */
   {                  0, SCANCODE_TAB              },	/* KEY_EDITOR_COLORS */

   {                  0, SCANCODE_1                },
   {                  0, SCANCODE_2                },
   {                  0, SCANCODE_3                },
   {                  0, SCANCODE_4                },
   {                  0, SCANCODE_5                },
   {                  0, SCANCODE_6                },
   {                  0, SCANCODE_7                },
   {                  0, SCANCODE_8                },
   {                  0, SCANCODE_9                },
   {                  0, SCANCODE_0                },
   {                  0, SCANCODE_R                },
   {                  0, SCANCODE_E                }
};

#endif /* USE_SVGALIB */

#ifdef USE_X11

KeySym x11_key_define[BD_MAXKEYS] =
{
   /***
    * game keys
    */
   XK_o,	/* KEY_GAME_UP      */
   XK_k,	/* KEY_GAME_DOWN    */
   XK_x,	/* KEY_GAME_LEFT    */
   XK_c,	/* KEY_GAME_RIGHT   */
   XK_0,	/* KEY_GAME_FIRE    */
   XK_q,	/* KEY_GAME_QUIT    */
   XK_s,	/* KEY_GAME_RESET   */
   XK_p,	/* KEY_GAME_PLAYERS */
   XK_t,	/* KEY_GAME_SCORE   */
   XK_e,	/* KEY_GAME_EDITOR  */

   /***
    * common keys
    */
   XK_h,	/* KEY_HELP         */
   XK_space,	/* KEY_SPACE        */
   XK_Return,	/* KEY_ENTER        */
   XK_F1,	/* KEY_SNAPSHOT     */

   /***
    * editor keys
    */
   XK_Up,	/* KEY_EDITOR_UP     */
   XK_Down,	/* KEY_EDITOR_DOWN   */
   XK_Left,	/* KEY_EDITOR_LEFT   */
   XK_Right,	/* KEY_EDITOR_RIGHT  */
   XK_C,	/* KEY_EDITOR_CLEAR  */
   XK_U,	/* KEY_EDITOR_UNDO   */
   XK_L,	/* KEY_EDITOR_LOAD   */
   XK_S,	/* KEY_EDITOR_SAVE   */
   XK_Q,	/* KEY_EDITOR_QUIT   */
   XK_Insert,	/* KEY_EDITOR_INSERT */
   XK_Home,	/* KEY_EDITOR_HOME   */
   XK_End,	/* KEY_EDITOR_END    */
   XK_Prior,	/* KEY_EDITOR_TOP    */
   XK_Next,	/* KEY_EDITOR_BOTTOM */
   XK_Tab,	/* KEY_GAME_EDITOR   */

   XK_1,
   XK_2,
   XK_3,
   XK_4,
   XK_5,
   XK_6,
   XK_7,
   XK_8,
   XK_9,
   XK_0,
   XK_r,
   XK_e
};

static ushort x11_key_table[128];


#ifdef ANSI_FUNC

static void key_x11_to_svga (KeySym keysym, short *key, short *key_shift)
#else
   
static void key_x11_to_svga (keysym, key, key_shift)
   KeySym keysym;
   short *key;
   short *key_shift;
#endif
{
   short i;

   *key_shift = 0;

   for (i = 0; i < 128; i++)
      if (keysym == keys_table[i].keysym)
      {
	 *key = i;

	 return;
      }
      else if (keysym == keys_table[i].keysym_shift)
      {
	 *key       = i;
	 *key_shift = 42;	/* left shift */

	 return;
      }
}


#ifdef ANSI_FUNC

static void key_update_x11_table (KeySym keysym, unchar status)
#else

static void key_update_x11_table (keysym, status)
   KeySym keysym;
   unchar status;
#endif
{
   short i;

   for (i = 0; i < 128; i++)
      if (keysym == keys_table[i].keysym)
      {
         x11_key_table[i] = status;

         return;
      }
      else if (keysym == keys_table[i].keysym_shift)
      {
         x11_key_table[42] = status;	/* left shift */
         x11_key_table[i]  = status;

         return;
      }
}


void key_callback (XKeyEvent *event)
{
   KeySym keysym;

   if (event->type == KeyPress)
   {
      (void) XLookupString (event, NULL, 0, &keysym, NULL);
      key_update_x11_table (keysym, 1);

      debug (("KeyPress: %04x\n", (int) keysym))
   }
   else
   {
      (void) XLookupString (event, NULL, 0, &keysym, NULL);
      key_update_x11_table (keysym, 0);

      debug (("KeyRelease: %04x\n", (int) keysym))
   }
}

#endif /* USE_X11 */


#ifdef ANSI_FUNC

short key_pressed_code (short key_code, short repeat_flag)
#else

short key_pressed_code (key_code, repeat_flag)
   short key_code;
   short repeat_flag;
#endif
{
   static short key_table[128];
   short        key = 0, key_shift;

   debug (("{%s(%d, %d, 0x%08x)", __FUNCTION__, key_code, repeat_flag))

#ifdef USE_SVGALIB
   if (game_data.graphics_type == GRAPHICS_SVGALIB)
   {
      if (key_code <= 0)
      {
         keyboard_clearstate ();
         (void) memset (key_table, 0, 128);

         debug_return ("} == %d", 0)
      }

      key_shift = svga_key_define[--key_code][0];
      key       = svga_key_define[key_code][1];

      debug (("key_code = %d, key = %d, key_shift = %d\n", key_code, key, key_shift))

      if (!key_shift || keyboard_keypressed (key_shift))
         if (keyboard_keypressed (key))
            if (repeat_flag || !key_table[key])
            {
               key_table[key] = 1;

               debug_return ("} == %d", 1)
            }
            else
               debug_return ("} == %d", 0)
   }
#endif

#ifdef USE_X11
   if (game_data.graphics_type == GRAPHICS_X11)
   {
      if (key_code <= 0)
      {
         (void) memset (key_table, 0, 128);
         (void) memset (x11_key_table, 0, 128);

         debug_return ("} == %d", 0)
      }

      key_x11_to_svga (x11_key_define[--key_code], &key, &key_shift);

      debug (("key_code = %d, key = %d, key_shift = %d\n", key_code, key, key_shift))

      if (!key_shift || x11_key_table[key_shift])
         if (x11_key_table[key])
	 {
            if (repeat_flag || !key_table[key])
            {
               key_table[key] = 1;

               debug_return ("} == %d", 1)
            }
            else
               debug_return ("} == %d", 0)
	 }
   }
#endif /* USE_X11 */

   key_table[key] = 0;

   debug_return ("} == %d", 0)
}


#ifdef ANSI_FUNC

short key_pressed_time (short key_code, short repeat_flag, time_t *current_time)
#else

short key_pressed_time (key_code, repeat_flag, current_time)
   short key_code;
   short repeat_flag;
   time_t *current_time;
#endif
{
   short retval;

   debug (("{%s(%d, %d, 0x%08lx)", __FUNCTION__, key_code, repeat_flag, (long) current_time))

   if ((retval = key_pressed_code (key_code, repeat_flag)))
      if (current_time)
         *current_time = time (NULL);

   debug_return ("} == %d", retval)
}


#ifdef ANSI_FUNC

short key_update (void)
#else

short key_update ()
#endif
{
#ifdef USE_SVGALIB
   if (game_data.graphics_type == GRAPHICS_SVGALIB)
      return (keyboard_update ());
#endif

#ifdef USE_X11
   if (game_data.graphics_type == GRAPHICS_X11)
      X11_handle_events ();
#endif /* USE_X11 */

   return (0);
}
